-- Christmas event


freeslot("SPR_XMAS")


local freezableFloors = {
	["LAKEFLR7"] = "NNZXMSF0",
	["JNGFLR12"] = "JNGXMSF0",
	["DCZFLR14"] = "ICEFLRA",
	["DCZFLR12"] = "ICEFLRA",
	["FWATER1" ] = "ICEFLRA",
	["LWATER1" ] = "ICEFLRA",
	["BWATER01"] = "ICEFLRA",
	["BRICK2"  ] = "NNZXMSF1",
	["THZFLR24"] = "THZXMSF0",
}

local freezableCeilings = {
	["LAKEFLR7"] = "NNZXMSF0",
	["JNGFLR12"] = "JNGXMSF0",
	["DCZFLR12"] = "ICEFLRA",
	["BRICK2"  ] = "NNZXMSF1",
}

local waterFlats = sf94.arrayToSet{
	"ICEFLRA", "FWATER1", "LWATER1",
	"BWATER01", "SURF01",
}

local snowlessFlats = sf94.arrayToSet{
	"F_SKY1"  , "PIT"     , "ICEFLRA",
	"NNZXMSF0", "JNGXMSF0", "NNZXMSF1",
	"GHZXMSF0", "NNZXMSF0", "FWATER1",
	"LWATER1" , "BWATER01", "SURF01",
}

local snowableMobjTypes = {}
for _, mtRange in ipairs{
	{ SPR_BUS1, B, MT_BERRYBUSH     },
	{ SPR_BUS2, B, MT_BUSH          },
	{ SPR_BUS3, B, MT_BLUEBERRYBUSH },

	{ SPR_XMAS, E, MT_GFZTREE      },
	{ SPR_XMAS, F, MT_GFZBERRYTREE },

	{ SPR_BSZ3, D, MT_BSZSHORTFLOWER_RED, MT_BSZSHORTFLOWER_ORANGE },
	{ SPR_BST4, A, MT_BSZTULIP_RED      , MT_BSZTULIP_ORANGE       },
	{ SPR_BSZ5, D, MT_BSZCLUSTER_RED    , MT_BSZCLUSTER_ORANGE     },
} do
	for mt = mtRange[3], mtRange[4] or mtRange[3] do
		snowableMobjTypes[mt] = { mtRange[1], mtRange[2] }

		-- NOTHINK prevents us from iterating through mobjs
		mobjinfo[mt].flags = $ & ~MF_NOTHINK
	end
end


local function freeze()
	for s in sectors.iterate do
		if not (sf94.event.noWaterFreeze
		or (gamemap == 1001 and s.tag >= 159 and s.tag <= 163)) then
			for fof in s.ffloors() do
				if waterFlats[fof.sector.ceilingpic] then
					fof.sector.ceilingpic = "ICEFLRA"
					fof.flags = $ & ~FF_SWIMMABLE | FF_SOLID & ~FF_RIPPLE
				end
			end
		end

		local frozenFloor = freezableFloors[s.floorpic]
		if frozenFloor then
			s.floorpic = frozenFloor
		end

		local frozenCeiling = freezableCeilings[s.ceilingpic]
		if frozenCeiling then
			s.ceilingpic = frozenCeiling
		end
	end

	for s in sectors.iterate do
		if s.ceilingpic ~= "F_SKY1" then continue end

		local top, topFof = s.floorheight, nil
		for fof in s.ffloors() do
			if fof.topheight > top and fof.flags & FF_RENDERPLANES then
				top, topFof = fof.topheight, fof
			end
		end

		local flat
		if topFof then
			flat = topFof.sector.ceilingpic
		else
			flat = s.floorpic
		end

		if flat == "GHZFLR07" then
			flat = "GHZXMSF0"
		elseif not snowlessFlats[flat] then
			flat = "SNOW01"
		end

		if topFof then
			topFof.sector.ceilingpic = flat
		else
			s.floorpic = flat
		end
	end

	for mo in mobjs.iterate() do
		local frozen = snowableMobjTypes[mo.type]
		if frozen then
			mo.sprite, mo.frame = frozen[1], frozen[2]
		end
	end
end

function sf94.startChristmasEvent()
	if not sf94.event then
		sf94.event = {}
	end
	sf94.event.id = "christmas"
	sf94.noChristmas = nil

	P_SwitchWeather(PRECIP_SNOW)
	sf94.nextWeatherChange = 0

	local map = G_BuildMapName(gamemap)

	if map == "MAPZ1" or map == "MAPZ2" or map == "MAPZ3" then
		freeze()
		P_LinedefExecute(1000)
		P_LinedefExecute(1100)
	elseif map ~= "MAPZ0" then
		freeze()
		S_ChangeMusic("sfxmas")
	end

	-- Remove the students from the classroom :)
	if map == "MAPZ2" then
		local removeList = {}
		for s in sectors.iterate do
			if s.tag == 1007 then
				for mo in s.thinglist() do
					if mo.type == MT_SF94_OLDSTANDNPC then
						table.insert(removeList, mo)
					end
				end
			end
		end
		for _, mo in pairs(removeList) do
			P_RemoveMobj(mo)
		end
	end

	-- Spawn tons of gifts :D
	if not sf94.event.noGift then
		sf94.spawnGifts()
	end
end

function sf94.stopChristmasEvent()
	sf94.event = nil
	sf94.nextWeatherChange = max(P_RandomRange(1, 4 * 24 * 3600 / 10) * 10 * TICRATE * 100 / 4, TICRATE * sf94.currentTimeFactor * sf94.timeFactor()) -- !!!
end

local function snowSpinTicker(mo)
	local snowball = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SF94_ROLLING_SNOWBALL)
	snowball.momx, snowball.momy = -1 * cos(mo.angle), -1 * sin(mo.angle)
end

function sf94.updateChristmasEvent()
	for p in players.iterate do
		local mo = p.mo
		if not (mo and mo.valid) then continue end

		if P_IsObjectOnGround(mo) then
			local ground
			if mo.z == mo.subsector.sector.floorheight then
				ground = mo.subsector.sector.floorpic
			else
				for fof in mo.subsector.sector.ffloors() do
					if fof.topheight == mo.z then
						ground = fof.sector.ceilingpic
						break
					end
				end
			end

			if ground == "ICEFLRA" then
				mo.friction = FU*127/128
			elseif p.pflags & PF_STARTDASH and p.dashspeed == p.maxdash
			and p.dashspeed ~= p.sf94.oldDashSpeed
			and ground == "SNOW01" then
				snowSpinTicker(mo)
			end
		end

		if p.mo.hat then
			sf94.updateSantaHat(p)
		end
	end
end


addHook("NetVars", function(n)
	sf94.event = n($)
	sf94.noChristmas = n($)
end)
