-- Events system


function sf94.updateEvent()
	if sf94.event.id == "christmas" then
		sf94.updateChristmasEvent()
	elseif sf94.event.id == "halloween" then
		sf94.updateHalloweenEvent()
	end
end

function sf94.startLocalEventStuff()
	if not sf94.event then return end

	local map = G_BuildMapName(gamemap)
	local mapSupported = (map == "MAPZ1" or map == "MAPZ2" or map == "MAPZ3")

	if sf94.event.id == "christmas" then
		if mapSupported then
			P_LinedefExecute(1100)
		elseif map ~= "MAPZ0" then
			S_ChangeMusic("sfxmas")
		end
	elseif sf94.event.id == "halloween" then
		if mapSupported then
			P_LinedefExecute(2100)
		end
	else
		return
	end
end


-- Tag: Linedef tag to be called if condition is verified
--
-- x-offset:
-- 0:  No event
-- 1:  Christmas
-- 2:  Halloween
-- -1: No event or Christmas
addHook("LinedefExecute", function(line, mo, sector)
	local id = line.frontside.textureoffset / FRACUNIT

	local trigger = false
	if sf94.event then
		if abs(id) == 1 and sf94.event.id == "christmas"
		or id == 2 and sf94.event.id == "halloween" then
			trigger = true
		end
	elseif id <= 0 then
		trigger = true
	end

	if trigger then
		P_LinedefExecute(line.tag, mo, sector)
	end
end, "CHECKEVENT")

COM_AddCommand("event", function(p, n, ...)
	if n then
		n = n:lower()
	else
		local c = "\x84"
		CONS_Printf(p, "\x82List of events:\x80")
		CONS_Printf(p, c.."event xmas\x80 (you can add \"nowaterfreeze\" (or \"nofreeze\") and \"nogift\")")
		CONS_Printf(p, c.."event halloween\x80 (the event will take effect on midnight)")
		CONS_Printf(p, c.."event stop/finish/end\x80 \n")
		return
	end

	if n == "stop" or n == "cancel" or n == "end" or n == "finish" then
		if sf94.event then
			if sf94.event.id == "christmas" then
				sf94.noChristmas = true
				CONS_Printf(p, "Christmas event will end on next map change.")
			elseif sf94.event.id == "halloween" then
				sf94.stopHalloweenEvent()
				CONS_Printf(p, "Halloween event stopped.")
			end
		else
			CONS_Printf(p, "There is no event to stop!")
		end
	elseif sf94.event then
		CONS_Printf(p, "Another event is currently active!\nType \"event stop\" to stop the current event.")
	elseif n == "christmas" or n == "xmas" then
		sf94.event = { id = "christmas" }
		for _, param in pairs({...}) do
			local s = param:lower()
			if s == "nofreeze" or s == "nowaterfreeze" then
				sf94.event.noWaterFreeze = true
			elseif s == "nogift" or s == "nogifts" then
				sf94.event.noGift = true
			else
				CONS_Printf(p, "\""..param.."\" is not a valid parameter!")
				sf94.event = nil
				return
			end
		end
		sf94.startChristmasEvent()
	elseif n == "halloween" then
		sf94.halloween = true
		CONS_Printf(p, "The Halloween event will be triggered when midnight comes.")
	else
		CONS_Printf(p, "\""..n.."\" is not a valid event name!\nType \"event\" with no parameters to show a list of all available events.")
	end
end, 1)
