-- Halloween event


freeslot(
	"MT_SF94_OLDNPC_ZOMBIE",
	"sfx_night",
	"sfx_npcz1",
	"sfx_blkout",
	"sfx_laugh1",
	"sfx_laugh2"
)


local zombiesSpeeds = { 20*FU, 12*FU, 16*FU, 8*FU }


local originalTriggerTags


local function zombieChase(mo)
	mo.tk, mo.fuse = 3, P_RandomRange(5*TICRATE, 15*TICRATE)
end

local function zombieTeleport(mo)
	local t = mo.target
	if not (mo.target and mo.target.valid) then
		mo.tk = 1
		mo.fuse = 3*TICRATE
		return
	end

	local x = t.x + P_RandomRange(-256, 256) * FU
	local y = t.y + P_RandomRange(-256, 256) * FU
	local z = t.z
	local s = R_PointInSubsector(x, y).sector
	if not P_CheckPosition(mo, x, y, z)
	or s.ceilingheight - s.floorheight < mo.height then
		return false
	end

	mo.tk = 4
	mo.fuse = 2*TICRATE
	P_SetOrigin(mo, x, y, z)
	mo.momx, mo.momy, mo.momz = 0, 0, 0
	for i = 1, 16 do
		local particle = P_SpawnGhostMobj(mo)
		particle.momx = P_RandomRange(-12, 12) * FU
		particle.momy = P_RandomRange(-12, 12) * FU
		particle.momz = P_RandomRange(-12, 12) * FU
	end

	return true
end

local function zombieQuickChase(mo)
	local t = mo.target
	if not (mo.target and mo.target.valid) then
		mo.tk = 1
		mo.fuse = 3*TICRATE
		return
	end
	local spd = zombiesSpeeds[mo.frame & FF_FRAMEMASK + 1] * 4

	mo.angle = R_PointToAngle2(mo.x, mo.y, t.x, t.y)
	P_InstaThrust(mo, mo.angle, spd)
	mo.tk, mo.fuse = 5, 3*TICRATE
end

local function none()
end

local zombiesThinkers = {

-- 1: Standing
none,

-- 2: Staring
function(mo)
	if not (mo.target and mo.target.valid) then
		mo.tk = 1
		mo.fuse = 3*TICRATE
		return
	end
	mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
end,

-- 3: Chasing
function(mo)
	local t = mo.target
	if not (t and t.valid and t.player and t.player.playerstate == PST_LIVE) then
		mo.tk = 1
		mo.fuse = 3*TICRATE
		return
	end

	local spd = zombiesSpeeds[mo.frame & FF_FRAMEMASK + 1]
	mo.angle = R_PointToAngle2(mo.x, mo.y, t.x, t.y)
	if R_PointToDist2(0, 0, mo.momx, mo.momy) < spd - 2*FU
	and P_IsObjectOnGround(mo) then
		mo.momz = 12*FU
	end
	P_InstaThrust(mo, mo.angle, spd)
end,

-- 4: Teleported
none,

-- 5: Quick chase
function(mo)
	P_SpawnGhostMobj(mo)
end

}

local function turnNPCIntoZombie(mo)
	mo.oldType, mo.oldColor = mo.type, mo.color

	if mo.type == MT_SF94_OLDNPC then
		mo.state = S_SF94_OLDNPC_ZOMBIE
	else
		mo.state = S_SF94_OLDNPC_ZOMBIE
		for i = 0, 2 do
			if mo.skin == skins[i].name then
				mo.frame = i
				break
			end
		end
	end
	mo.frame = $ | FF_FULLBRIGHT | TR_TRANS50
	mo.color = SKINCOLOR_WHITE

	mo.type = MT_SF94_OLDNPC_ZOMBIE
	mo.flags = $ & ~(MF_SOLID | MF_PUSHABLE | MF_SHOOTABLE) | MF_SPECIAL | MF_SLIDEME

	if P_LookForPlayers(mo) then
		mo.tk, mo.fuse = 2, P_RandomRange(1, 15*TICRATE)
	else
		mo.tk, mo.fuse = 1, P_RandomRange(1, 3*TICRATE)
	end
end

local function turnZombieIntoNPC(mo)
	mo.frame = $ & ~(FF_FULLBRIGHT | TR_TRANS50)
	mo.type, mo.oldType = $2, nil
	if mo.type == MT_SF94_OLDNPC then
		mo.state = S_SF94_OLDNPC_SPAWN
		mo.flags = $ | MF_SOLID | MF_PUSHABLE | MF_SHOOTABLE & ~(MF_SPECIAL | MF_SLIDEME)
	else
		mo.state = S_SF94_OLDSTANDNPC

		local sp = mo.spawnpoint
		P_SetOrigin(mo, sp.x * FU, sp.y * FU, sp.z * FU)
		mo.angle = sp.angle * FU / 360 * FU
		mo.momx, mo.momy, mo.momz = 0, 0, 0
		mo.flags = $ | MF_SOLID & ~MF_PAIN & ~MF_SPECIAL & ~MF_SLIDEME
	end
	mo.color = mo.oldColor
	mo.oldColor = nil

	mo.fuse = 0
	mo.tk = nil
end

function sf94.startHalloweenEvent()
	sf94.event = {
		id = "halloween",
		strokes = 12,
		nextStroke = leveltime,
		nextSound = leveltime + P_RandomRange(30, 120) * TICRATE
	}
	sf94.halloween = nil

	local map = G_BuildMapName(gamemap)
	if not sf94.tagsToDisableForHalloween[map] then return end

	if VERSIONSTRING == "v2.2.10" then
		for _, tags in ipairs(sf94.tagsToDisableForHalloween[map]) do
			for _, s in ipairs(sf94.getSectorsByTags(tags)) do
				s.special = 0
			end
		end
	else
		originalTriggerTags = {}

		for _, tags in ipairs(sf94.tagsToDisableForHalloween[map]) do
			for _, s in ipairs(sf94.getSectorsByTags(tags)) do
				originalTriggerTags[s] = s.triggertag
				s.triggertag = 0
			end
		end
	end

	if map == "MAPZ3" then
		S_StartSound(nil, sfx_fizzle)
		sf94.event.nextStroke = $ + 6*TICRATE
	end

	for mo in mobjs.iterate() do
		if mo.type >= MT_SF94_OLDNPC and mo.type <= MT_SF94_OLDSTANDNPC_SKIN4 then
			turnNPCIntoZombie(mo)
		elseif mo.type >= MT_SF94_EXCLAMATIONMARK and mo.type <= MT_SF94_DOTS then
			mo.flags2 = $ | MF2_DONTDRAW
		elseif mo.type == MT_SF94_SONICSTATUE then
			mo.frame = 1
		end
	end

	P_LinedefExecute(2000)
	P_LinedefExecute(2100)
end

function sf94.stopHalloweenEvent()
	sf94.event = nil

	local map = G_BuildMapName(gamemap)
	if not sf94.tagsToDisableForHalloween[map] then return end

	if VERSIONSTRING == "v2.2.10" then
		for _, tags in ipairs(sf94.tagsToDisableForHalloween[map]) do
			for _, s in ipairs(sf94.getSectorsByTags(tags)) do
				s.special = tags.effect
			end
		end
	else
		if originalTriggerTags then
			for _, tags in ipairs(sf94.tagsToDisableForHalloween[map]) do
				for _, s in ipairs(sf94.getSectorsByTags(tags)) do
					s.triggertag = originalTriggerTags[s]
				end
			end

			originalTriggerTags = nil
		end
	end

	if map == "MAPZ3" then
		for i, s in ipairs(sf94.lightSectors) do
			s.lightlevel = sf94.lights[i]
		end
		sf94.setPreviousLight(255)
	end

	for mo in mobjs.iterate() do
		if mo.type == MT_SF94_OLDNPC_ZOMBIE then
			turnZombieIntoNPC(mo)
		elseif mo.type >= MT_SF94_EXCLAMATIONMARK and mo.type <= MT_SF94_DOTS then
			mo.flags2 = $ & ~MF2_DONTDRAW
		elseif mo.type == MT_SF94_SONICSTATUE then
			mo.frame = 0
		end
	end

	P_LinedefExecute(2500)
end

function sf94.updateHalloweenEvent()
	local ev = sf94.event

	if sf94.light > 224 or sf94.time > 1200000 then
		sf94.stopHalloweenEvent()
	end

	if ev.strokes ~= 0 and leveltime >= ev.nextStroke then
		S_StartSound(nil, sfx_night)
		ev.nextStroke = leveltime + 90
		ev.strokes = $ - 1
	end

	if leveltime >= ev.nextSound then
		ev.nextSound = leveltime + P_RandomRange(10, 120) * TICRATE

		local s = P_RandomByte()
		if s > 63 then
			s = sf94.randomElement{
				sfx_ambin2, sfx_doorc1, sfx_doord1,
				sfx_rumbam, sfx_rumble, sfx_timeup,
				sfx_laugh1, sfx_laugh2, sfx_blkout
			}
		else
			s = sf94.randomElement{
				sfx_s3kbdl, sfx_s3k59 , sfx_s3k5e ,
				sfx_s3k9b , sfx_wbreak, sfx_becrsh,
				sfx_beelec
			}
		end
		S_StartSound(nil, s)
	end
end


addHook("MobjThinker", function(mo)
	zombiesThinkers[mo.tk](mo)
end, MT_SF94_OLDNPC_ZOMBIE)

addHook("MobjFuse", function(mo)
	if mo.tk == 1 then -- Standing
		if P_LookForPlayers(mo) then
			mo.tk, mo.fuse = 2, P_RandomRange(1, 15*TICRATE)
		else
			mo.fuse = P_RandomRange(1, 3*TICRATE)
		end

	elseif mo.tk == 2 then -- Staring
		zombieChase(mo)

	elseif mo.tk == 3 then -- Chasing
		if P_RandomByte() > 63 then
			if P_RandomByte() > 63 then
				zombieQuickChase(mo)
			else
				if not zombieTeleport(mo) then
					mo.fuse = P_RandomRange(3*TICRATE, 10*TICRATE)
				end
			end
		else
			local t = mo.target
			if not (t and t.valid
			and (P_AproxDistance(t.x - mo.x, t.y, mo.y) <= 4096*FU
			or P_CheckSight(mo, t)
			or not P_LookForPlayers(mo)))
				mo.tk = 2
				mo.fuse = P_RandomRange(1, 15*TICRATE)
			else
				mo.fuse = P_RandomRange(5*TICRATE, 15*TICRATE)
			end
		end

	elseif mo.tk == 4 then -- Teleported
		if P_RandomByte() > 63 then
			zombieChase(mo)
		else
			zombieQuickChase(mo)
		end

	elseif mo.tk == 5 then -- Quick chase
		zombieChase(mo)

	end

	return true
end, MT_SF94_OLDNPC_ZOMBIE)

addHook("TouchSpecial", function(mo, pmo)
	local p = pmo.player
	if not (p and p.valid) then return true end

	P_InstaThrust(mo, P_RandomKey(FU) * FU, 48*FU)
	mo.target = pmo
	mo.fuse = 3*TICRATE

	if mo.tk == 4 then return true end
	mo.tk = 2

	p.losstime = max($, 120*TICRATE)
	if P_DamageMobj(pmo, mo, mo) then
		S_StartSound(mo, sfx_npcz1)
	end

	return true
end, MT_SF94_OLDNPC_ZOMBIE)

addHook("MapChange", function()
	originalTriggerTags = nil
end)

addHook("NetVars", function(n)
	sf94.halloween = n($)
	originalTriggerTags = n($)
end)
