-- Day/night cycle and weather changing by LJ Sonic (SF94-Serv version)


local exceptions = { "MAP16", "MAP22", "MAP23", "MAP24", "MAP25", "MAP55", "MAP56", "MAP57", "MAPF4", "MAPM4", "MAPM5", "MAPM6", "MAPZ0", "MAPZ3" }
local rainExceptions = { "MAP13", "MAP16", "MAP22", "MAP23", "MAP24", "MAP25", "MAP54", "MAP55", "MAP56", "MAP57", "MAPF4", "MAPM2", "MAPM5", "MAPM6" }
local snowExceptions = { "MAP13", "MAP16", "MAP22", "MAP23", "MAP24", "MAP25", "MAP54", "MAP55", "MAP56", "MAP57", "MAPF4", "MAPF6", "MAPM2", "MAPM4", "MAPM5", "MAPM6" }
local stormExceptions = { "MAP22", "MAP23", "MAP24", "MAP25", "MAP56", "MAP57", "MAPF4", "MAPM6" }


local MINUTE = 60*TICRATE
local HOUR = 60*MINUTE
local DAY = 24*HOUR


local baseLight = 0
local previousLight = 255
local sky = 1
local previousSky = 0
local refresh = true

local active = CV_RegisterVar({"lightchange", "Auto", CV_NETVAR, {On = 1, Off = 0, Auto = 2}})
local timeFactor = CV_RegisterVar({"timefactor", 60, CV_NETVAR, {MIN = 0, MAX = DAY}})
local nightTimeFactor = CV_RegisterVar({"nighttimefactor", 16, CV_NETVAR, {MIN = 1, MAX = DAY}})
local daySky = CV_RegisterVar({"daysky", 0, CV_CALL | CV_NETVAR, CV_Unsigned, function() refresh = true end})
local nightSky = CV_RegisterVar({"nightsky", 9, CV_CALL | CV_NETVAR, CV_Natural, function() refresh = true end})
local morningSky = CV_RegisterVar({"morningsky", 79, CV_CALL | CV_NETVAR, CV_Natural, function() refresh = true end})
local eveningSky = CV_RegisterVar({"eveningsky", 78, CV_CALL | CV_NETVAR, CV_Natural, function() refresh = true end})
local clock = CV_RegisterVar({"clock", "On", 0, CV_OnOff})
local randomWeather = CV_RegisterVar({"randomweather", "On", CV_NETVAR, {On = 1, Off = 0, Rain = 2, Snow = 3, NoThunder = 4, RainNoThunder = 5, NoRain = 6, Thunder = 7}})
local randomWeatherFactor = CV_RegisterVar({"randomweatherfactor", 100, CV_NETVAR, {MIN = 1, MAX = 10000}})


-- Helper function to bypass the limited range of P_RandomRange
local function randomRange(a, b)
	local x = P_RandomRange(a / 100, b / 100) * 100
	return min(max(x, a), b)
end

local function calculateLight()
	if sf94.event and sf94.event.id == "halloween"
	and G_BuildMapName(gamemap) == "MAPZ3" then
		baseLight = 144
		sf94.light = 144
		sky = 80
		return true
	end

	if active.value == 0 or active.value == 2 and sf94.lightException then
		baseLight = 255
	else
		if sf94.time > 19*HOUR then
			if sf94.time < 23*HOUR then
				baseLight = 255 - (255 - 144) * (sf94.time - 19*HOUR) / ((23 - 19) * HOUR)
			else
				baseLight = 144
			end
		elseif sf94.time < 9*HOUR then
			if sf94.time > 5*HOUR then
				baseLight = 144 + (255 - 144) * (sf94.time - 5*HOUR) / ((9 - 5) * HOUR)
			else
				baseLight = 144
			end
		else
			baseLight = 255
		end
	end

	if baseLight <= 160 then
		sky = nightSky.value
	elseif baseLight > 224 then
		if daySky.value == 0 then
			sky = mapheaderinfo[gamemap].skynum
		else
			sky = daySky.value
		end
	elseif sf94.time <= 9*HOUR then
		sky = morningSky.value
	elseif sf94.time >= 19*HOUR then
		sky = eveningSky.value
	else
		sky = daySky.value
	end

	if curWeather == PRECIP_SNOW or (sf94.event and sf94.event.id == "christmas") then
		if sf94.weatherLight < 255 then
			sf94.weatherLight = $ + 1
		end
		sky = 77
	elseif sf94.event and sf94.event.id == "halloween" then
		sky = 80
	elseif curWeather ~= PRECIP_NONE and curWeather ~= PRECIP_BLANK then
		if sf94.weatherLight < 192 then
			sf94.weatherLight = $ + 1
		end
		if sf94.weatherLight > 192 then
			sf94.weatherLight = $ - 1
		end
		if baseLight > 160 then
			sky = 20
		end
	elseif sf94.weatherLight < 255 then
		sf94.weatherLight = $ + 1
	end

	sf94.light = baseLight * sf94.weatherLight / 255

	-- Don't let the map become too dark during bad weather
	if sf94.light < 144 then
		sf94.light = max($, baseLight)
	end

	return active.value ~= 0 and (active.value ~= 2 or not sf94.lightException)
end

local function checkVertice(i, x, y)
	return vertexes[i].x == x * FRACUNIT and vertexes[i].y == y * FRACUNIT
end

local function initLight()
	sf94.weatherLight, sf94.lights, sf94.lightSectors = 255, {}, {}

	for sector in sectors.iterate do
		table.insert(sf94.lightSectors, sector)
		table.insert(sf94.lights, sector.lightlevel)
	end

	sf94.lightException = false
	local mapName = G_BuildMapName(gamemap)
	for _, rejectedMap in ipairs(exceptions) do
		if mapName == rejectedMap then
			sf94.lightException = true
			break
		end
	end

	if calculateLight() then
		if sf94.event and sf94.event.id == "halloween"
		and G_BuildMapName(gamemap) == "Z3" then
			for sector in sectors.iterate do
				sector.lightlevel = min($, baseLight)
			end
		else
			for sector in sectors.iterate do
				sector.lightlevel = $ * baseLight / 255
			end
		end

		previousLight = baseLight
	else
		previousLight = 255
	end
	P_SetupLevelSky(sky)

	-- Used to be a special case for SRB2 Hub. Too lazy to update it for 2.2.
	--[[if #lines ~= 5462 return end
	if #sides ~= 10203 then return end
	if #sectors ~= 831 then return end
	if not checkVertice(1100, -7584, -11552) then return end
	if not checkVertice(2200, -11264, -3712) then return end
	if not checkVertice(3300, -13056, -11904) then return end
	if not checkVertice(4400, -9440, -15904) then return end

	for i, sector in ipairs({3, 46, 50, 54, 95, 101, 103, 112, 121, 122, 148, 168, 184, 191, 193, 195, 208, 218, 219, 226, 231, 256, 264, 282, 347, 375, 376, 384, 436, 488, 491, 492, 493, 494, 495, 500, 510, 514, 524, 555, 557, 564, 584, 592, 595, 599, 613, 615, 617, 618, 631, 649, 654, 656, 659, 661, 666, 668, 670, 675, 678, 690, 725, 732, 733, 739, 743, 744, 749, 753, 755, 756, 757, 759, 762, 763, 764, 765, 766, 767, 768, 769, 770, 771, 780, 782, 787, 790, 791, 793, 794, 795, 796, 797, 798, 799, 800, 801, 806, 807, 808, 809, 811, 824}) do
		table.remove(sf94.lightSectors, sector - i + 2)
		table.remove(sf94.lights, sector - i + 2)
	end]]
end

local function applyLight()
	local lightSectors = sf94.lightSectors
	local lights = sf94.lights
	local light = sf94.light

	--if sf94.event and sf94.event.id == "halloween"
	--and G_BuildMapName(gamemap) == "Z3" then

	for i = 1, #lightSectors do
		local s = lightSectors[i]
		local l = lights[i]

		if l > light then
			s.lightlevel = light
		else
			s.lightlevel = l
		end
	end

	--for i = 1, #lightSectors do
	--	local s = lightSectors[i]
	--	local l = lights[i]
	--
	--	s.lightlevel = l * light / 255
	--end

	previousLight = sf94.light
end

local function changeWeather()
	local weather = PRECIP_BLANK
	local duration = randomRange(1, 4 * DAY / TICRATE)
	local tries = 16

	if randomWeather.value then
		repeat
			if curWeather == PRECIP_RAIN then
				if P_RandomKey(256) <= 127 then
					weather = PRECIP_STORM
					duration = randomRange(1, DAY / TICRATE / 2)
				else
					weather = PRECIP_BLANK
				end
			elseif curWeather == PRECIP_STORM or curWeather == PRECIP_STORM_NOSTRIKES then
				if P_RandomKey(256) <= 127 then
					weather = PRECIP_RAIN
				else
					weather = PRECIP_STORM_NORAIN
				end
				duration = randomRange(1, DAY / TICRATE / 2)
			elseif curWeather == PRECIP_STORM_NORAIN then
				if P_RandomKey(256) <= 127 then
					weather = PRECIP_STORM
					duration = randomRange(1, DAY / TICRATE / 2)
				else
					weather = PRECIP_BLANK
				end
			elseif curWeather == PRECIP_SNOW then
				weather = PRECIP_BLANK
			elseif curWeather == PRECIP_NONE or curWeather == PRECIP_BLANK then
				weather = P_RandomKey(256)
				if weather <= 191 then
					if weather <= 95 then
						weather = PRECIP_RAIN
						if P_RandomKey(256) <= 127 then
							duration = randomRange(1, DAY / TICRATE)
						else
							duration = randomRange(1, 2 * HOUR)
						end
					else
						weather = PRECIP_STORM_NORAIN
						if P_RandomKey(256) <= 127 then
							duration = randomRange(1, DAY / TICRATE / 2)
						else
							duration = randomRange(1, 2 * HOUR)
						end
					end
				else
					weather = PRECIP_SNOW
					duration = randomRange(1, 2 * DAY / TICRATE)
				end
			end

			tries = $ - 1
		until weather == PRECIP_NONE or weather == PRECIP_BLANK or weather == PRECIP_RAIN and randomWeather.value ~= 3 and randomWeather.value < 6
		or weather == PRECIP_SNOW and randomWeather.value ~= 2 and randomWeather.value ~= 5 and randomWeather.value ~= 7
		or weather == PRECIP_STORM and randomWeather.value < 3
		or weather == PRECIP_STORM_NORAIN and (randomWeather.value < 3 or randomWeather.value > 5) or tries == 0

		if weather == PRECIP_STORM or weather == PRECIP_STORM_NORAIN or weather == PRECIP_STORM_NOSTRIKES then
			local mapName = G_BuildMapName(gamemap)
			for _, rejectedMap in ipairs(stormExceptions) do
				if mapName == rejectedMap then
					if weather == PRECIP_NORAIN then
						weather = PRECIP_BLANK
					else
						weather = PRECIP_RAIN
					end
					break
				end
			end
		end
		if weather == PRECIP_RAIN or weather == PRECIP_STORM or weather == PRECIP_STORM_NOSTRIKES then
			local mapName = G_BuildMapName(gamemap)
			for _, rejectedMap in ipairs(rainExceptions) do
				if mapName == rejectedMap then
					if weather == PRECIP_RAIN then
						weather = PRECIP_BLANK
					else
						weather = PRECIP_STORM_NORAIN
					end
					break
				end
			end
		elseif weather == PRECIP_SNOW then
			local mapName = G_BuildMapName(gamemap)
			for _, rejectedMap in ipairs(snowExceptions) do
				if mapName == rejectedMap then
					weather = PRECIP_BLANK
					break
				end
			end
		end

		P_SwitchWeather(weather)
	end

	sf94.nextWeatherChange = max($ + duration * TICRATE * 100 / 4, TICRATE * sf94.currentTimeFactor * timeFactor.value)
end

function sf94.setPreviousLight(n)
	previousLight = n
end

function sf94.timeFactor()
	return timeFactor.value
end


COM_AddCommand("displaytime", function(p)
	CONS_Printf(p, ("It is %d:%.2d."):format(
		sf94.time / HOUR,
		sf94.time / MINUTE % 60
	))
end)

COM_AddCommand("time", function(p, hours, minutes, seconds)
	if hours then
		hours = hours:lower()
	end

	if hours == "day" then
		sf94.time = 9*HOUR
		return
	elseif hours == "night" then
		sf94.time = 23*HOUR
		return
	elseif hours == "morning" then
		sf94.time = 5*HOUR
		return
	elseif hours == "evening" then
		sf94.time = 19*HOUR
		return
	end

	hours = tonumber($)
	if hours == nil then
		CONS_Printf(p, "Syntax: time <hours> [minutes [seconds]]\n        time {day|night|morning|evening}")
		return
	end
	if hours < 0 or hours > 23 then
		CONS_Printf(p, "Number of hours has to range between 0 and 23.")
		return
	end

	minutes = tonumber($)
	if minutes == nil then
		minutes, seconds = 0, 0
	end
	if minutes < 0 or minutes > 59 then
		CONS_Printf(p, "Number of minutes has to range between 0 and 59.")
		return
	end

	seconds = tonumber($)
	if seconds == nil then
		seconds = 0
	end
	if seconds < 0 or seconds > 59 then
		CONS_Printf(p, "Number of seconds has to range between 0 and 59.")
		return
	end

	sf94.time = HOUR * hours + MINUTE * minutes + TICRATE * seconds
end, COM_ADMIN)

COM_AddCommand("weather", function(p, weather, duration)
	if duration ~= "infinite" then
		duration = tonumber(duration)
		if duration ~= nil then
			if duration <= 0 then
				duration = "infinite"
			elseif duration < 17044 then
				duration = $ * HOUR
			else
				CONS_Printf(p, "Please choose a shorter duration!")
			end
		end
	end

	if weather then
		weather = weather:lower()
	end

	if weather == "none"
	or weather == "clear"
	or weather == "sun" then
		weather = PRECIP_BLANK
		duration = $ or 2 * DAY
	elseif weather == "storm" then
		weather = PRECIP_STORM
		duration = $ or DAY / 4
	elseif weather == "snow" then
		weather = PRECIP_SNOW
		duration = $ or DAY
	elseif weather == "rain" then
		weather = PRECIP_RAIN
		duration = $ or DAY / 2
	elseif weather == "thunder" then
		weather = PRECIP_STORM_NORAIN
		duration = $ or DAY / 4
	elseif weather == "stormnostrikes" then
		weather = PRECIP_STORM_NOSTRIKES
		duration = $ or DAY / 4
	else
		CONS_Printf(p, 'Syntax: weather {sun|rain|snow|storm|thunder|stormnostrikes} [duration]')
		return
	end

	if duration == "infinite" then
		sf94.nextWeatherChange = 0
	else
		sf94.nextWeatherChange = duration * 100 / 4
	end

	P_SwitchWeather(weather)
end, COM_ADMIN)

local function help(p)
	if not (p and p.valid) then return end

	CONS_Printf(p, "\x83\nCommands\x80")
	CONS_Printf(p, "time displaytime weather")
	CONS_Printf(p, "\x83\nVariables\x80")
	CONS_Printf(p, "lightchange timefactor nighttimefactor daysky nightsky morningsky eveningsky clock randomweather randomweatherfactor\n")
end

COM_AddCommand("timehelp", help)
COM_AddCommand("helptime", help)

addHook("ThinkFrame", function()
	if gamestate ~= GS_LEVEL then return end

	if sf94.time == nil then
		sf94.time = 8 * HOUR
		sf94.currentTimeFactor = 1
		sf94.light = 255
		sf94.nextWeatherChange = randomRange(1, 4 * DAY / TICRATE) * (TICRATE * 100 / 4)
		sf94.weatherLight = 255
		sf94.lightException = false
		previousSky = 0
	end
	--if sf94.lights == nil initLight() end -- Apparently crashy in netgames?

	sf94.currentTimeFactor = 1

	if sf94.nextWeatherChange ~= 0 then
		sf94.nextWeatherChange = $ - timeFactor.value * sf94.currentTimeFactor * randomWeatherFactor.value
		if sf94.nextWeatherChange <= 0 then
			changeWeather()
		end
	end

	if sf94.halloween and sf94.time < timeFactor.value * sf94.currentTimeFactor then
		sf94.startHalloweenEvent()
	end

	sf94.time = ($ + timeFactor.value * sf94.currentTimeFactor) % DAY

	if sky ~= previousSky then
		P_SetupLevelSky(sky)
		previousSky = sky
	end

	if sf94.lightSectors == nil then return end
	if not calculateLight() then return end
	if sf94.light ~= previousLight then
		refresh = true
	end
	if not refresh then return end
	refresh = false

	applyLight()
end)

addHook("MapChange", do
	sf94.lightSectors = nil
	sf94.lightException = false
	previousSky = 0
end)

addHook("MapLoad", do
	initLight()
end)

addHook("NetVars", function(n)
	sf94.time = n($)
	sf94.currentTimeFactor = n($)

	sf94.light = n($)
	sf94.lights = n($)
	sf94.lightSectors = n($)
	sf94.lightException = n($)

	sf94.nextWeatherChange = n($)
	sf94.weatherLight = n($)
end)

hud.add(function(v, p)
	if sf94.time == nil then return end

	if clock.value then
		local s = ("%d:%.2d"):format(sf94.time / HOUR, sf94.time / MINUTE % 60)
		v.drawString(316, 190, s, V_SNAPTOBOTTOM | V_SNAPTORIGHT, "right")
	end
end)
